"""Top-level package for Global Logger."""

__author__ = """Alexey Rubasheff"""
__email__ = 'alexey.rubasheff@gmail.com'
__version__ = '0.3.26'

from .global_logger import Log   # noqa # pylint: disable=unused-import
